
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#include "dl_list.h"
#include "fdf.h"

extern int errno;

#define BUF_SIZE 1024

#define DUMMY_SYMLINK_FD (FILE *)(-1)

struct centry_s {
	struct centry_s *child, *parent;
	fentry_s *fe;
	FILE *fd;
	unsigned char buf[BUF_SIZE];
	ssize_t read;
	off_t first_diff;
	bool eof;
};

// TODO: baum wie vorher überlegt, aber zwischenspeichern des letzten überprüften integers

static struct centry_s *buckets[256];

static void output(struct centry_s *v) {
	struct centry_s *ce;
	bool pfound[2] = { false, false };
	fentry_s *fe;
	
	start_output();
	
	for (ce = v; ce != NULL; ce = ce->child)
		pfound[ce->fe->flags.preferred] = true;
	if (pfound[false] && pfound[true]) {
		for (ce = v; ce != NULL; ce = ce->child)
			if (ce->fe->flags.preferred)
				output_fe(ce->fe, true);
	} else {
		for (ce = v; ce != NULL; ce = ce->child)
			output_fe(ce->fe, false);
	}
	
	stop_output();
}

static bool centry_close(struct centry_s *v, int from) {
	bool r = true;
	DBG("!!! ======= closing %s [ from: %d ]\n", v->fe->path, from);
	if (v->fd != DUMMY_SYMLINK_FD && fclose(v->fd)) {
		if (opts.failfast)
			r = false;
		ERROR(_("error closing file '%s': %s\n"), v->fe->path, strerror(errno));
	}
	v->fd = NULL;
	return r;
}

static void narrow_list(dlst_t l, off_t off, ssize_t max) {
	int i;
	dlst_entry_t e;
	struct centry_s *v;
	
	DBG("narrowing list (size: %d)\n", dlst_size(l));
	
	for (i=0; i<max && !dlst_empty(l); i++) {
		dlst_forall(e, l) {
			v = dlst_data(l, e);
			v->child = v->parent = NULL;
		}
		
		DBG("round %d of %d... (to test: %d)\n", i, max, dlst_size(l));
		memset(buckets, 0x00, sizeof(buckets));
		e = dlst_first(l);
		while (e != NULL) {
			v = dlst_data(l, e);
			if (v->read < i) {
				centry_close(v, 1);
				dlst_entry_t en = dlst_next(l, e);
				dlst_remove(l, e);
				e = en;
				continue;
			}
			unsigned char d = v->buf[i];
			DBG("read d: %d for %s\n", d, v->fe->path);
			if (buckets[d] != NULL) {
				DBG("adding double '%s' to '%s'\n", v->fe->path, buckets[d]->fe->path);
				v->child = buckets[d];
				buckets[d]->parent = v;
			}
			buckets[d] = v;
			
			e = dlst_next(l, e);
		}
		
		e = dlst_first(l);
		while (e != NULL) {
			v = dlst_data(l, e);
			dlst_entry_t en = dlst_next(l, e);
			if (!v->child && !v->parent) {
				MSG("found single file: %s\n", v->fe->path);
				centry_close(v, 2);
				v->first_diff = off + i;
				dlst_remove(l, e);
			}
			e = en;
		}
	}
}

bool check_list(dlst_t l, off_t size) {
	const size_t len = dlst_size(l);
	struct centry_s *es = calloc(len, sizeof(struct centry_s));
	unsigned int i = 0;
	dlst_t to;
	dlst_init(&to);
	bool r = true;
	
	struct centry_s *ed = NULL;
	dlst_entry_t e = dlst_first(l);
	while (e != NULL) {
		fentry_s *fe = dlst_data(l, e);
		
		es[i].child = ed;
		if (ed != NULL)
			ed->parent = &(es[i]);
		ed = &(es[i]);
		
		if (S_ISLNK(fe->mode)) {
			es[i].fd = DUMMY_SYMLINK_FD;
		} else {
			es[i].fd = fopen(fe->path, "r");
		}
		
		if (!es[i].fd) {
			ERROR(_("error opening file '%s': %s\n"), fe->path, strerror(errno));
			if (opts.failfast) {
				r = false;
				goto end;
			}
			dlst_entry_t en = dlst_next(l, e);
			dlst_remove(l, e);
			e = en;
			continue;
		}
		
		es[i].fe = fe;
		DBG("pushing for %s: 0x%08x (child's parent: 0x%08x, child: 0x%08x)\n", fe->path, (unsigned int)&es[i],
				es[i].child ? (unsigned int)es[i].child->parent : -1U, (unsigned int)es[i].child);
		dlst_push(to, &(es[i]));
		i++;
		e = dlst_next(l, e);
	}
	
	if (opts.msglevel == DEBUG) {
		for (e = dlst_first(to); e != NULL; e = dlst_next(to, e)) {
			struct centry_s *ce = dlst_data(to, e);
			printf("pushed for %s: 0x%08x (parent: 0x%08x, child: 0x%08x)\n", ce->fe->path, (unsigned int)ce,
					(unsigned int)ce->parent, (unsigned int)ce->child);
		}
	}
	
	off_t done = 0;
	dlst_entry_t de, en;
	while (done < size && dlst_size(to) > 0 && !interrupted) {
		ssize_t max_read = 0;
		de = dlst_first(to);
		while (de != NULL) {
			DBG("next: %s (%llu bytes)\n",
					((struct centry_s *)dlst_data(to, de))->fe->path,
					((struct centry_s *)dlst_data(to, de))->fe->size);
			struct centry_s *v = dlst_data(to, de);
			FILE *fd = v->fd;
			if (!fd) {
				de = dlst_next(to, de);
				continue;
			}
			if (v->eof)	{ // e < rest, insert cmp()->(!cmp())+
				r &= centry_close(v, 3);
				goto error;
			}
			if (fd == DUMMY_SYMLINK_FD) {
				ssize_t len;
				char *ln_path = readlink_malloc(v->fe->path, &len);
				if (!ln_path) {
					ERROR(_("error reading symlink '%s': %s\n"), v->fe->path, strerror(errno));
					goto error;
				}
				v->read = len - done;
				strncpy((char *)v->buf, ln_path, len);
				free(ln_path);
				v->eof = (v->read < BUF_SIZE);
			} else {
				DBG("exec read(0x%08x, %s, %u, %d, %d)\n",
						(unsigned int)v->buf, v->fe->path, sizeof(char), BUF_SIZE, (int)fd);
				v->read = fread(v->buf, sizeof(char), BUF_SIZE, fd);
				DBG("done reading (read %d bytes)\n", v->read);
				if (v->read < BUF_SIZE) {
					v->eof = feof(fd);
					int err = ferror(fd);
					if (err) {
						ERROR("error reading '%s': %s\n", v->fe->path, strerror(err));
						r = false;
						centry_close(v, 4);
						goto error;
					}
				}
			}
			max_read = max(max_read, v->read);
			de = dlst_next(to, de);
			continue;
			
error:
			en = dlst_next(to, de);
			dlst_remove(to, de);
			de = en;
			
		}
		size_t before = dlst_size(to);
		narrow_list(to, done, max_read);
		DBG("after narrowing: list size %d -> %d, data done: %llu -> %llu / %llu\n",
				before, dlst_size(to), done, done + max_read, size);
		done += max_read;
	}
	
	if (!interrupted)
		dlst_forall(de, to)
			if (((struct centry_s *)dlst_data(to, de))->parent == NULL) {
				DBG("starting output: %s\n", ((struct centry_s *)dlst_data(to, de))->fe->path);
				output(dlst_data(to, de));
			}
	
end:
	dlst_forall(de, to)
		r &= centry_close(dlst_data(to, de), 5);
	dlst_clear(to);
	free(es);
	return r;
}
